! $Id: conv_objects.f90,v 1.2 2001/04/20 09:48:55 salam Exp $

!----------------------------------------------------------------------
! This is useful in a couple of disconnected places, but should probably
! not be made available except to a few internal routines.
module pdf_representation
  use types; use consts_dp
  !----------------------------------------------------------------------
  integer, parameter, public :: ncomponents=2
  integer, parameter, public :: id_g=0, id_u=1, id_d=2
end module pdf_representation







!======================================================================
! Contains objects representing and dealing with actual convolutions.
! So it will contain the splitting and coefficient function holders and 
! the routines for using them in a convolution. The routines themselves
! will assume the following representation for the splitting function
!        /  g      \
!    V = |  up     |
!        \  down   /
!
! where up actually means all quarks with charge +-2/3
!     down                                       +-1/3
!
! THIS MODULE SHOULD NOT BE USEd: access its components through 
! conv_objects (necessary for reasons to do with 
!======================================================================
module conv_objects_hidden
  use types; use consts_dp; use splitting_functions
  use qcd
  use pdf_representation
  use convolution
  use assertions
  implicit none
  private

  !-----------------------------------------------------------------
  ! holds all the components of the splitting functions
  type PMat
     !private
     !-- in analogy with LO, these are the non-singlet (i.e. P_V) when 
     !   working at NLO (note different notation from Bryan).
     !   Note that these are intended to be splitting to a single quark
     !   species, not the whole bunch of quarks and anti-quarks
     type(grid_conv)    :: gg, qq, gq, qg
     !-- these are the additional pieces which appear at NLO
     type(grid_conv)    :: qqbarV, qqS
     logical            :: NLO
  end type PMat
  public             :: PMat


  !-----------------------------------------------------------------
  ! holds the components of a coefficient function
  type CMat
     !private
     !-- need a grid def. Sometimes no g or q will be defined
     !   but still want the coefficient function associated with a grid
     !   def
     !   delta is the magnitude of a delta function piece
     !   quark charge contributions are included by the routines here.
     !   for gluons the implicitness is \sum_{q,qbar} e_q^2
     type(grid_def)     :: gd
     type(grid_conv)    :: g, q
     real(dp)           :: delta
     logical            :: HO
  end type CMat
  public             :: CMat


  !---------- avoid need for main program to access convolution
  !  give it both possible names. Maybe not necessary?
  public :: grid_def
  interface cobj_InitGridDef
     module procedure conv_InitGridDef
  end interface
  public :: cobj_InitGridDef
  public :: conv_InitGridDef

  public :: cobj_InitSplitLO, cobj_PConv, cobj_InitSplitNLO
  public :: cobj_InitSplit, cobj_AddSplit, cobj_DelSplit

  !-------- things for splitting functions --------------------
  interface cobj_InitCoeff
     module procedure cobj_InitCoeffLO, cobj_InitCoeffHO, cobj_InitCoeff_cf
  end interface
  public :: cobj_InitCoeff, cobj_InitCoeffLO, cobj_InitCoeffHO
  public :: cobj_CConv
  public :: cobj_AddCoeff, cobj_DelCoeff

contains
  !----------------------------------------------------------------------
  subroutine cobj_InitSplitLO(gd, P)
    type(grid_def), intent(in)    :: gd
    type(PMat),     intent(inout) :: P
    if (nf_d /= (nf_int+1)/2) write(0,*) 'WARNING: non-standard nf_d'
    call conv_InitGridConv(gd, P%gg, sf_Pgg)
    call conv_InitGridConv(gd, P%qq, sf_Pqq)
    call conv_InitGridConv(gd, P%gq, sf_Pgq)
    call conv_InitGridConv(gd, P%qg, sf_Pqg)
    P%NLO = .false.
  end subroutine cobj_InitSplitLO

  !----------------------------------------------------------------------
  subroutine cobj_InitSplitNLO(gd, P, factscheme)
    type(grid_def), intent(in)    :: gd
    type(PMat),     intent(inout) :: P
    integer, optional, intent(in) :: factscheme
    integer :: factscheme_local
    if (nf_d /= (nf_int+1)/2) write(0,*) 'WARNING: non-standard nf_d'

    if (present(factscheme)) then
       factscheme_local = factscheme
    else
       factscheme_local = factscheme_default
    end if

    select case (factscheme_local)
    case (factscheme_MSbar)
       call conv_InitGridConv(gd, P%qq, sf_P1qqV)
       call conv_InitGridConv(gd, P%qg, sf_P1qg)
    case (factscheme_DIS)
       write(0,*) 'DIS scheme splitting functions not supported &
            &from cobj_InitSplitNLO'
       stop
       !call conv_InitGridConv(gd, P%qq, sf_P1qqV_DIS)
       !call conv_InitGridConv(gd, P%qg, sf_P1qg_DIS)
    case default
       write(0,*) 'cobj_InitSplitNLO: unsupported fact scheme', factscheme
    end select

    call conv_InitGridConv(gd, P%gq, sf_P1gq)
    call conv_InitGridConv(gd, P%gg, sf_P1gg)
    call conv_InitGridConv(gd, P%qqbarV, sf_P1qqbarV)
    call conv_InitGridConv(gd, P%qqS, sf_P1qqS)
    P%NLO = .true.
  end subroutine cobj_InitSplitNLO


  !----------------------------------------------------------------------
  ! init a splitting function set with another one (potentially 
  ! multiplied by some factor)
  subroutine cobj_InitSplit(P, Pin, fact)
    type(PMat),     intent(inout) :: P
    type(PMat),     intent(in)    :: Pin
    real(dp),       intent(in), optional :: fact
    if (nf_d /= (nf_int+1)/2) write(0,*) 'WARNING: non-standard nf_d'

    P%NLO = Pin%NLO
    call conv_InitGridConv(P%gg,    Pin%gg,     fact)
    call conv_InitGridConv(P%qq,    Pin%qq,     fact)
    call conv_InitGridConv(P%gq,    Pin%gq,     fact)
    call conv_InitGridConv(P%qg,    Pin%qg,     fact)
    if (P%NLO) then
       call conv_InitGridConv(P%qqbarV,Pin%qqbarV, fact)
       call conv_InitGridConv(P%qqS,   Pin%qqS,    fact)
!!$    else
!!$       call conv_DelGridConv(P%qqbarV)
!!$       call conv_DelGridConv(P%qqS)
    end if
  end subroutine cobj_InitSplit

  !----------------------------------------------------------------------
  ! init a splitting function set with another one (potentially 
  ! multiplied by some factor)
  subroutine cobj_AddSplit(P, Padd, fact)
    type(PMat),     intent(inout) :: P
    type(PMat),     intent(in)    :: Padd
    real(dp),       intent(in), optional :: fact
    if (nf_d /= (nf_int+1)/2) write(0,*) 'WARNING: non-standard nf_d'

    call conv_AddGridConv(P%gg,    Padd%gg,     fact)
    call conv_AddGridConv(P%qq,    Padd%qq,     fact)
    call conv_AddGridConv(P%gq,    Padd%gq,     fact)
    call conv_AddGridConv(P%qg,    Padd%qg,     fact)
    if (P%NLO .and. Padd%NLO) then
       call conv_AddGridConv(P%qqbarV,Padd%qqbarV, fact)
       call conv_AddGridConv(P%qqS,   Padd%qqS,    fact)
    else if ((.not. P%NLO) .and. Padd%NLO) then
       call conv_InitGridConv(P%qqbarV,Padd%qqbarV, fact)
       call conv_InitGridConv(P%qqS,   Padd%qqS,    fact)
    end if
    P%NLO = P%NLO .or. Padd%NLO
  end subroutine cobj_AddSplit

  !----------------------------------------------------------------------
  ! Clear a splitting function (dealloc memory)
  subroutine cobj_DelSplit(P)
    type(PMat),     intent(inout) :: P

    call conv_DelGridConv(P%gg)
    call conv_DelGridConv(P%qq)
    call conv_DelGridConv(P%gq)
    call conv_DelGridConv(P%qg)
    if (P%NLO) then
       call conv_DelGridConv(P%qqbarV)
       call conv_DelGridConv(P%qqS   )
    end if
  end subroutine cobj_DelSplit

  !----------------------------------------------------------------------
  ! This will be overloaded...
  function cobj_PConv(P,q) result(Pxq)
    type(PMat), intent(in) :: P
    real(dp),   intent(in) :: q(0:,0:)
    real(dp)               :: Pxq(0:ubound(q,dim=1),0:ncomponents)
    !--- need this temporary array
    real(dp) :: PqgG(0:ubound(q,dim=1)), PqqSq(0:ubound(q,dim=1))

    Pxq(:,id_g) = P%gg * q(:,id_g) + P%gq * (q(:,id_u)+q(:,id_d))
    PqgG(:)     = P%qg * q(:,id_g)
    Pxq(:,id_u) = P%qq * q(:,id_u) + two*nf_u * PqgG(:)
    Pxq(:,id_d) = P%qq * q(:,id_d) + two*nf_d * PqgG(:)

    if (P%NLO) then
       PqqSq = P%qqS * (q(:,id_u) + q(:,id_d))
       Pxq(:,id_u) = Pxq(:,id_u) + (P%qqbarV * q(:,id_u)) + two*nf_u * PqqSq(:)
       Pxq(:,id_d) = Pxq(:,id_d) + (P%qqbarV * q(:,id_d)) + two*nf_d * PqqSq(:)
    end if
    !Pxq(:,id_u) = PqgG(:)
  end function cobj_PConv

  !======================================================================
  ! From here onwards we have things to do with coefficient functions
  subroutine cobj_InitCoeffLO(gd,C, fact)
    type(grid_def), intent(in)    :: gd
    type(Cmat),     intent(inout) :: C
    real(dp),       intent(in), optional :: fact

    C%gd = gd
    C%HO = .false.
    if (present(fact)) then
       C%delta = fact
    else
       C%delta = one
    end if
  end subroutine cobj_InitCoeffLO
  
  !-- HO. coeff function is not quite so simple. Note that we do simply 
  !   call it NLO, because it could also be relevant at O(as^2).
  !   NOTE UNUSUAL ORDER...
  subroutine cobj_InitCoeffHO(gd, C, coeff_g, coeff_q)
    use convolution_communicator
    type(grid_def), intent(in)    :: gd
    type(Cmat),     intent(inout) :: C
    interface
       function coeff_g(x)
         use types; implicit none
         real(dp), intent(in) :: x
         real(dp)             :: coeff_g
       end function coeff_g
    end interface
    interface
       function coeff_q(x)
         use types; implicit none
         real(dp), intent(in) :: x
         real(dp)             :: coeff_q
       end function coeff_q
    end interface
    real(dp) :: sanity1, sanity2

    C%gd = gd
    C%HO = .true.
    C%delta = zero
    call conv_InitGridConv(gd,C%g, coeff_g)
    call conv_InitGridConv(gd,C%q, coeff_q)
    !-- a sanity check: quite often the arguments for the
    !   quark and gluon pieces get exchanged.
    !   Cannot always ensure correctness (e.g. FL), but can 
    !   detect some cases of misuse
    cc_piece = cc_VIRT
    sanity1 = coeff_g(one)
    cc_piece = cc_DELTA
    sanity2 = coeff_g(zero)
    if (sanity1 /= zero .or. sanity2 /= zero) then
       write(0,*) 'WARNING in cobj_InitCoeffHO **********************'
       write(0,*) 'gluon coefficient function has virtual corrections'
       write(0,*) 'this could be a sign that quark and gluon cf fns have&
            & been exchanged'
       write(0,*) '**************************************************'
    end if
  end subroutine cobj_InitCoeffHO

  !-- initialise C as being fact*Cin ----------------------
  subroutine cobj_InitCoeff_cf(C, Cin, fact)
    type(Cmat),     intent(inout) :: C
    type(Cmat),     intent(in)    :: Cin
    real(dp),       intent(in), optional :: fact

    C%gd = Cin%gd
    C%HO = Cin%HO
    C%delta = Cin%delta
    if (present(fact)) C%delta = C%delta * fact
    if (C%HO) then
       call conv_InitGridConv(C%q, Cin%q, fact)
       call conv_InitGridConv(C%g, Cin%g, fact)
    end if
  end subroutine cobj_InitCoeff_cf

  
  !-- initialise C as being fact*Cin ----------------------
  subroutine cobj_AddCoeff(C, Cin, fact)
    type(Cmat),     intent(inout) :: C
    type(Cmat),     intent(in)    :: Cin
    real(dp),       intent(in), optional :: fact

    call conv_ValidateGD(C%gd,Cin%gd,'cobj_AddCoeff')
    if (present(fact)) then 
       C%delta = C%delta  + Cin%delta * fact
    else
       C%delta = C%delta  + Cin%delta
    end if
    
    if (Cin%HO) then
       if (C%HO) then
          call conv_AddGridConv(C%q, Cin%q, fact)
          call conv_AddGridConv(C%g, Cin%g, fact)
       else
          call conv_InitGridConv(C%q, Cin%q, fact)
          call conv_InitGridConv(C%g, Cin%g, fact)
       end if
    end if
    C%HO = C%HO .or. Cin%HO
  end subroutine cobj_AddCoeff

  !----------------------------------------------------------------------
  ! Clear a coefficient function (dealloc memory)
  subroutine cobj_DelCoeff(C)
    type(CMat),     intent(inout) :: C

    if (C%HO) then
       call conv_DelGridConv(C%g)
       call conv_DelGridConv(C%q)
    end if
  end subroutine cobj_DelCoeff



  !-------------------------------------------------------------------
  function cobj_CConv(C,q) result(Cxq)
    type(CMat), intent(in) :: C
    real(dp),   intent(in) :: q(0:,0:)
    real(dp)               :: Cxq(0:ubound(q,dim=1))
    !--------------------------------------------------
    real(dp) :: ud(0:ubound(q,dim=1))
    real(dp), parameter :: uq2 = (4.0_dp/9.0_dp)
    real(dp), parameter :: dq2 = (1.0_dp/9.0_dp)

    ud = q(:,id_u)*uq2 + q(:,id_d)*dq2
    if (C%HO) then
       !-- CHECK: factor of two in front of nf_u and nf_d is there both 
       !   in ESW and the ADS paper.
       Cxq = (two*(nf_u*uq2 + nf_d*dq2)) * (C%g .conv. q(:,id_g))&
            & + (C%q .conv. ud)
       if (C%delta /= zero) Cxq = Cxq + C%delta * ud
    else
       Cxq = C%delta * ud
    end if
  end function cobj_CConv
    
end module conv_objects_hidden


!======================================================================
! In the above module it was not possible to overload * and .conv. for
! ev_PConv. The reason for this is not immediately clear, but it may be
! connected to the fact that these already existed (and had been included
! through a use specification) and the public statement did not make clear
! whether it referred to the included operators or the newly defined ones. 
!
! A solution is to not use evolution_hidden, but evolution, which includes
! everything that is public in evolution_hidden, but not those things that
! were accessible only there (notably operators included from
! convolution). Then here we define the new additional operators.
!
! NOTE: exceptionally this module is not private.
module conv_objects
  use conv_objects_hidden
  implicit none


  interface operator(*)
     module procedure cobj_PConv, cobj_CConv
  end interface
  public :: operator(*)
  interface operator(.conv.)
     module procedure cobj_PConv, cobj_CConv
  end interface
  public  :: operator(.conv.)
  public  :: cobj_Eval2LConv
  !-- avoid access through here
  private :: cobj_PConv, cobj_CConv


contains
  !-------------------------------------------------------------------
  ! Calculates [(C2 - y^2/(1+(1-y)^2) CL) .conv. pdf] .atx. xbj
  ! put it here to have access to .conv. for coefficient functions
  function cobj_Eval2LConv(C2, CL, pdf, xbj, ybj) result(res)
    use types; use consts_dp; use convolution
    type(CMat), intent(in)  :: C2, CL
    real(dp),   intent(in)  :: pdf(0:,0:), xbj, ybj
    real(dp)                :: res
    !----------------------------------------------
    type(gdval) :: gdx
    type(Cmat)  :: C2L

    call cobj_InitCoeff(C2L, C2)
    call cobj_AddCoeff (C2L, CL, - ybj**2/(1+(1-ybj)**2) )
    gdx = xbj .with. C2%gd
    res = conv_EvalGridQuant(gdx%gd, (C2L .conv. pdf), -log(gdx%val))
    !res = ((C2L.conv.pdf) .atx. gdx ) 
    call cobj_DelCoeff (C2L)
  end function cobj_Eval2LConv

end module conv_objects


