! $Id: event_shapes.f90,v 1.23 2002/02/19 20:41:41 salam Exp $

!======================================================================
! This contains all the information describing the event shapes?
!
! All coefficient functions are returned in the DIS scheme
! To get MSbar scheme results, one must add C2_1 for the MSbar scheme
!
!
! Nomenclature: tau_zQ means tau wrt to the z axis, normalised to Q.
!               rho_tE
!
!
! CHECKLIST IF ADDING NEW PARAMETERS OR PROPERTIES
!  
!   . update nparam, or nprop                      xx
!   . define integer value for handle              xx
!   . add pointers in type evshp_info              xx
!   . initialise pointers in es_ResummedCoeffs     xx
!   . initialise values in dynamic_initiate        xx
!
! 
! CHECKLIST IF ADDING NEW EVENT SHAPES
!   . many things not written here
!   . update event_shape_ids
!
! GPS 13/06/2000
!======================================================================
module event_shapes
  use types; use consts_dp
  use qcd; use warnings_and_errors
  implicit none
  private

  !-- all the non-functional info one might want about an event shape
  type evshp_info
     real(dp) :: G12, G11, G23, G22
     !-- GP means that we have q( var**GP * Q^2)
     real(dp) :: GP
     !-- coefficient for resummation off a gluon
     real(dp) :: GCA12
     !-- coefficient corresponding to the non-global piece
     !  (already included in G22, but sometimes it is to be subtracted out)
     real(dp) :: G22_NG
     !-- sub-leading pieces that enter only when doing a NON-STANDARD
     !   resummation. They are coefficients of asb^2, asb^2 L
     !   and of asb^2 P.conv.q, asb^2 L * P .conv. q
     !   There are two varieties for two possible non-standard situations.
     real(dp) :: ns20(2), ns21(2), ns20_GP(2), ns21_GP(2)
     !-- above coeffs apply to log(var/rescale)
     real(dp) :: rescale 
     !-- power correction coefficient (multiplies \cP)
     real(dp) :: c_V
     !--- put these at the end to keep nice alignment?
     logical  :: resumCA
     integer  :: esid
     !--- from here onwards put things that the user may want
     !    to be able to update dynamically?
     real(dp), pointer :: esmax, esPTModPower, esNPModPower, esRescaleV
     integer,  pointer :: esSumType, esC1MuF
  end type evshp_info
  public :: evshp_info

  !------- the ids for the different event shapes  ------
  ! Remember to update the "public" definitions of event_shape_ids if 
  ! we add something to this list
  integer, parameter, public :: EvshpId_illegal = -1
  integer, parameter, public :: EvshpId_tau_zQ = 1
  integer, parameter, public :: EvshpId_tau_zE = 2
  integer, parameter, public :: EvshpId_tau_tQ = 3
  integer, parameter, public :: EvshpId_tau_tE = 4
  integer, parameter, public :: EvshpId_brd_zE = 5
  integer, parameter, public :: EvshpId_brd_tE = 6
  integer, parameter, public :: EvshpId_rho_E  = 7
  integer, parameter, public :: EvshpId_Cpr_E  = 8
  integer, parameter, public :: EvshpId_CrntEQ = 9  ! current energy norm Q
  integer, parameter, public :: EvshpId_Cempty = 10
  integer, parameter, public :: EvshpId_rho23E = 11
  integer, parameter :: n_esid = 11

  real(dp), parameter, public :: es_undefined_value  = -1e90_dp
  integer,  parameter, public :: es_undefined_intval = -1000000034

  !------- the names for the different event shapes ------
  character(len=6), parameter, public :: es_names(n_esid) = (/&
       &'tau_zQ',&
       &'tau_zE',&
       &'tau_tQ',&
       &'tau_tE',&
       &'brd_zE',&
       &'brd_tE',&
       &'rho_E ',&
       &'Cpr_E ',&
       &'CrntEQ',&
       &'Cempty',&
       &'rho23E'/)     
  
  integer, parameter :: es_comp_2q = 1
  integer, parameter :: es_comp_2g = 2
  integer, parameter :: es_comp_Lq = 3
  integer, parameter :: es_comp_Lg = 4
  integer, parameter :: es_comp_CA2q = 5
  integer, parameter :: es_comp_CALq = 6
  integer, parameter :: es_comp_CA2g = -1  ! should never be called
  integer, parameter :: es_comp_CALg = -2  ! should never be called

  !-------------------------------------------------------
  ! fixed order properties that may be updated dynamically
  !-------------------------------------------------------
  logical          :: dynamic_initiated = .false.
  !real(dp), target :: dynamic_esmax(n_esid)
  !integer,  target :: dynamic_esSumType(n_esid)
  !------------ new try -------------
  ! this must be public for automatic include-file generarion
  integer, parameter, public :: es_nparam = 4
  integer, parameter :: nparam = es_nparam, nprop = 2
  integer, parameter :: propbase=nparam
  !-- these are real numbers
  real(dp), target :: dynamic_param(nparam, n_esid)
  !-- these are integer numbers
  integer,  target :: dynamic_prop(propbase+1:propbase+nprop, n_esid)

  !-- these are the handles for the PARAMeters
  integer, parameter, public :: esMax = 1
  integer, parameter, public :: esPTModPower = 2
  integer, parameter, public :: esNPModPower = 3
  integer, parameter, public :: esRescaleV   = 4

  !-- these are the handles for the PROPerties
  integer, parameter, public :: esSumType = es_nparam+1
  integer, parameter, public ::      esSumType_standard    = 0
  integer, parameter, public ::      esSumType_WithLbar_m  = 1
  integer, parameter, public ::      esSumType_WithLbar    = 2
  integer, parameter, public ::      esSumType_NoNonGlobal = 3
  integer, parameter, public :: esC1MuF = es_nparam+2
  integer, parameter, public ::      esC1MuF_Q  = 0
  integer, parameter, public ::      esC1MuF_VQ = 1
  !-- event shape id
  integer :: es_id = EvshpId_illegal
  public :: es_setid, es_idOfName
  public :: es_C2q, es_C2g, es_CLq, es_CLg
  public :: es_CCA2q, es_CCA2g, es_CCALq, es_CCALg
  public :: es_ResummedCoeffs
  !-- these can be changed by hand; but not very good practice!
  public :: es_UpdateDynamicESMax, es_GetDynamicESMax
  public :: es_UpdateDynamicESSumType, es_GetDynamicESSumType
  public :: es_UpdateDynamicParam, es_GetDynamicParam
  public :: es_UpdateDynamicProp, es_GetDynamicProp
  !-- for my personal use!
  interface es_UpdateDynamic
     module procedure es_UpdateDynamicParam, es_UpdateDynamicProp
  end interface
  interface es_GetDynamic
     module procedure es_GetDynamicParam, es_GetDynamicProp
  end interface
  public :: es_UpdateDynamic, es_GetDynamic
  
contains

  !--------------------------------------------------
  ! These are the sole entry points for this module
  ! All actual work will be dealt with by internal routines

  !--------------------------------------------------
  ! sets the id of the event shape to be used
  subroutine es_setid(id)
    integer, intent(in) :: id
    es_id = id
  end subroutine es_setid
  

  !--------------------------------------------------------------
  ! returns the id corresponding to a given name, or EvshpId_illegal
  ! if the name is not supported.
  function es_idOfName(name) result(id)
    character(len=*) :: name
    integer :: id
    do id = 1, n_esid
       if (name == trim(es_names(id))) then
          return
       end if
    end do
    id = EvshpId_illegal
  end function es_idOfName
  

  !--------------------------------------------------------------------
  ! the routines for the convolution functions for the main resummation
  !----------------------------------------------------------------------

  ! For vars with resumn off gluon, this will be C_q, as defined by Mrinal
  ! and so in use in front of the CF resummation we will need to include
  ! C_q - CCAq
  function es_C2q(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = es_Cgeneric(y, es_comp_2q)
  end function es_C2q
  function es_C2g(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = es_Cgeneric(y, es_comp_2g)
  end function es_C2g
  ! See above note about vars with CA resumn
  function es_CLq(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = es_Cgeneric(y, es_comp_Lq)
  end function es_CLq
  function es_CLg(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = es_Cgeneric(y, es_comp_Lg)
  end function es_CLg
  !------------------------------------------------------
  ! in the situation that we need resummation off a gluon
  ! See above note about vars with CA resumn
  ! Corresponds to Mrinals C_q^g
  function es_CCA2q(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = es_Cgeneric(y, es_comp_CA2q)
  end function es_CCA2q
  ! See above note about vars with CA resumn
  function es_CCALq(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = es_Cgeneric(y, es_comp_CALq)
  end function es_CCALq
  function es_CCA2g(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = zero   ! this piece is always zero
  end function es_CCA2g
  function es_CCALg(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = zero   ! likewise
  end function es_CCALg
  
  
  !----------------------------------------------------------------------
  ! sets up the coefficients describing the resummation
  ! Most things have their usual meaning.
  ! GP means that we have q( var**GP * Q^2)
  subroutine es_ResummedCoeffs(es_info)
    type(evshp_info), intent(out) :: es_info
    real(dp), parameter :: brd_zE_X12 = 1.945031318_dp
    real(dp)            :: LambdaBar_20, LambdaBar_21, LambdaBar_22
    real(dp)            :: LambdaBar_20_g, LambdaBar_21_g
    real(dp)            :: LambdaBar_m_21, LambdaBar_m_22 
    !-------------------------------

    !-- set non-standard pieces to zero... They can then be reset
    !   individually
    es_info%ns20 = zero
    es_info%ns21 = zero
    es_info%ns20_GP = zero
    es_info%ns21_GP = zero
    es_info%G22_NG = zero

    select case (es_id)
    case (EvshpId_tau_zQ)
       es_info%G12 = -2*CF
       es_info%G11 = 3*CF
       es_info%G23 = twopi*beta0*es_info%G12
       es_info%G22 = -4/3.0_dp*pisq*CF**2 + (pisq/three - 169/36.0_dp)*CA*CF&
                  & + 11/18.0_dp*CF*nf*two*tr
       es_info%GP  = one
       es_info%resumCA = .false.
       es_info%GCA12 = +1e200_dp  ! make things go crazy if this gets used?
       es_info%rescale = one
       es_info%c_V     = two
    case (EvshpId_tau_zE)
       es_info%G12 = -1.5_dp*CF
       es_info%G11 = 3*(1-ln2)*CF
       es_info%G23 = 16*pi/9.0_dp * beta0*es_info%G12
!!$       !- original version
!!$       es_info%G22 = -2/3.0_dp*pisq*CF**2 + (pisq/four - 17/6.0_dp - 22/3.0_dp&
!!$                & *ln2)*CA*CF + (one/three + four/three*ln2)*CF*nf*two*tr
       !-- new version
       es_info%G22 = -2/3.0_dp*pisq*CF**2 + (pisq/four - 17/6.0_dp - 22/3.0_dp&
                & *ln2)*CA*CF + (one/three + four/three*ln2)*CF*nf*two*tr&
            & - cf * ca * pisq/12.0_dp
       es_info%G22_NG = - cf * ca * pisq/12.0_dp
       es_info%GP  = one
       es_info%resumCA = .false.
       es_info%GCA12 = +1e200_dp  ! make things go crazy if this gets used?
       es_info%rescale = one
       es_info%c_V     = two
    case (EvshpId_rho_E, EvshpId_Cpr_E, EvshpId_tau_tE)
       !-- just  half of those for EvshpId_tau_zQ? NO: not in case of G22...
       es_info%G12 = -CF
       es_info%G11 = 3*half*CF
       es_info%G23 = twopi*beta0*es_info%G12
       !-- both of the following results seem to give correct answers
       !   but need to add the piece from double large-angle soft...
       es_info%G22 = - cf / 6.0_dp * ( -9*pi*beta0 + 2*cf*pisq + 6*cmw_K )&
            & - cf * ca * pisq/three
       es_info%G22_NG = - cf * ca * pisq/3.0_dp
       es_info%GP  = zero
       !-- nice, special event-shapes
       es_info%resumCA = .true.
       es_info%GCA12 = es_info%G12*CA/CF  
       select case(es_id)
       case(EvshpId_rho_E)
          es_info%rescale = one
          es_info%c_V     = one
       case(EvshpId_Cpr_E)
          es_info%rescale = 12.0_dp
          es_info%c_V     = three*pi
       case(EvshpId_tau_tE)
          es_info%rescale = 2.0_dp
          es_info%c_V     = two
       end select
    case(EvshpId_brd_zE)
       es_info%G12 = -4*CF
       !es_info%G11 = 0
       !write(0,*) 'NB: FAKE G11 for brd_E'
       es_info%G11 = (6 - 4*ln2)*CF
       es_info%G23 = -32/three * pi * beta0 * CF
       es_info%G22 = ((12 - 16*ln2)*pi*beta0 - 4*cmw_K &
            &         - (two/three*pisq + 24*ln2**2)*CF)*CF
       es_info%GP  = two
       es_info%resumCA = .false.
       es_info%GCA12 = +1e200_dp  ! make things go crazy if this gets used?
       es_info%rescale = one
       es_info%c_V     = one
       !--- non-standard pieces: not always used, but want them available
       ! Obtained from CCN20-31, 54, 61
       LambdaBar_20 = three/10.0_dp*CF**2*pisq**2& ! standard piece
            & - 8*CF*zeta3*beta0*pi &! this piece is associated with R'''
            & +(four*eulergamma-three)*CF**2*(3*zeta3) !R2gp in Lmbdbar
       LambdaBar_21 = four/three*CF*(&
            & -two*pi**3*beta0+9.0_dp*CF*zeta3+6*CF*brd_zE_X12)&
            & +(four*eulergamma-three)&
            &     *CF**2*(-one+7*pisq/6.0_dp-two*ln2**2) !R2gp in Lmbdbar
       LambdaBar_22 = two/three*(7*pisq-36*ln2**2-6.0_dp)*CF**2
       !-- coefficients of gamma in LambdaBar from R2gp
       LambdaBar_20_g = (two)*CF*(3*zeta3)
       LambdaBar_21_g = (two)*CF*(-one+7*pisq/6.0_dp-two*ln2**2)
       !-- these are needed (CCN20-65) depending on choice for Rpeff
       LambdaBar_m_22 = (half*pisq - 4)*CF**2
       LambdaBar_m_21 = -pi**3*beta0*CF

       !-- for the following refer to CCN20-65
       es_info%ns20(ESSumType_WithLbar_m) = &
         & LambdaBar_20 - eulergamma*LambdaBar_m_21
       es_info%ns21(ESSumType_WithLbar_m) = &
         & LambdaBar_21 - 3*CF**2 - eulergamma*(2*LambdaBar_m_22 + 4*CF**2)
       es_info%ns20(ESSumType_WithLbar) = &
         & LambdaBar_20 - eulergamma*LambdaBar_21
       es_info%ns21(ESSumType_WithLbar) = &
         & LambdaBar_21 - 3*CF**2 - eulergamma*(2*LambdaBar_22 + 4*CF**2)
       !-- these are the same in the two schemes
       es_info%ns21_GP = LambdaBar_21_g + 2*CF
       es_info%ns20_GP = LambdaBar_20_g
       !-- tmp for testing
!!$       es_info%ns20(:) = LambdaBar_20
!!$       es_info%ns21(:) = LambdaBar_21 - 3*CF**2
!!$       es_info%ns21_GP = 2*CF
!!$       es_info%ns20_GP = zero




!!$       es_info%ns20 = three/10.0_dp*CF**2*pisq**2&
!!$            & - 8*CF*zeta3*beta0*pi &! this piece is associated with R'''
!!$            & +(-three*CF)*CF*(3*zeta3) !R2gp in Lmbdbar
!!$       es_info%ns21 = four/three*CF*(&
!!$         & -two*pi**3*beta0+9.0_dp*CF*zeta3+6*CF*brd_zE_X12) - three*CF**2&
!!$         & +(-three*CF)*CF*(-one+7*pisq/6.0_dp-two*ln2**2) !R2gp in Lmbdbar
!!$       ! last line from inclusion of R2gp in Lambdabar
!!$       es_info%ns21_GP = two*CF !-two*CF ! <-- similarly

    case (EvshpId_rho23E)
       !-- just  half of those for EvshpId_tau_zQ? NO: not in case of G22...
       es_info%G12 = zero
       es_info%G11 = zero
       es_info%G23 = zero
       !-- both of the following results seem to give correct answers
       es_info%G22 = zero
       es_info%GP  = zero
       !-- nice, special event-shapes
       es_info%resumCA = .true.
       es_info%GCA12 = -CA
       es_info%rescale = one
       es_info%c_V     = es_undefined_value
    case (EvshpId_CEmpty)
       !-- a sort of fake event shape: no resummation supported
       !   but allows one to use it easily with match holder
       es_info%G12 = zero; es_info%G11 = zero; es_info%G23 = zero; 
       es_info%G22 = zero; es_info%GP  = zero
       es_info%resumCA = .false.; es_info%GCA12 = zero
       es_info%rescale = one
       es_info%c_V     = es_undefined_value
    case default
       write(0,*) 'ERROR. es_ResumCoeffs: unsupported event-shape&
            & variable: ', es_id
       stop
    end select
    !-- set this once we've checked there are no problems?
    es_info%esid = es_id
    !-- now the dynamic stuff
    if (.not. dynamic_initiated) call dynamic_initiate
    !--- parameters ----------------------------------
    es_info%esmax        => dynamic_param(esMax, es_id)
    es_info%esPTModPower => dynamic_param(esPTModPower, es_id)
    es_info%esNPModPower => dynamic_param(esNPModPower, es_id)
    es_info%esRescaleV   => dynamic_param(esRescaleV,   es_id)
    !--- properties ----------------------------------
    es_info%esSumType => dynamic_prop(esSumType, es_id)
    es_info%esC1MuF   => dynamic_prop(esC1MuF, es_id)
  end subroutine es_ResummedCoeffs
  

  !----------------------------------------------------------------------
  ! The dynamic stuff: things that the user may want to update during the
  ! program running and which should apply to all invocations of the 
  ! given event-shape.
  !
  ! From one point of view this is not very nice, because it means that
  ! info about event shapes is being defined in two separate places
  subroutine dynamic_initiate
    !-- do it only once!
    if (dynamic_initiated) return
    dynamic_initiated = .true.

    !-- for the ones I forget to define
!!$    dynamic_esmax(:) = es_undefined_value
!!$    dynamic_esmax(EvshpId_tau_zQ) = one
!!$    dynamic_esmax(EvshpId_tau_zE) = one
!!$    dynamic_esmax(EvshpId_tau_tE) = half
!!$    dynamic_esmax(EvshpId_rho_E ) = 0.25_dp
!!$    dynamic_esmax(EvshpId_Cpr_E ) = one
!!$    dynamic_esmax(EvshpId_brd_zE) = half
!!$
!!$    dynamic_esSumType(:) = ESSumType_standard
!!$    dynamic_esSumType(EvshpId_brd_zE) = ESSumType_WithLbar_m

    !--- PARAMS -----------------
    dynamic_param(esMax, :) = es_undefined_value
    dynamic_param(esMax, EvshpId_tau_zQ) = one
    dynamic_param(esMax, EvshpId_tau_zE) = one
    dynamic_param(esMax, EvshpId_tau_tE) = half
    dynamic_param(esMax, EvshpId_rho_E ) = 0.25_dp
    dynamic_param(esMax, EvshpId_Cpr_E ) = one
    dynamic_param(esMax, EvshpId_brd_zE) = half
    !-- PT modification power ------------
    dynamic_param(esPTModPower, : ) = one
    !-- NP modification power ------------
    dynamic_param(esNPModPower, : ) = one
    !-- rescaling of observable before resummation
    dynamic_param(esRescaleV,   : ) = one
    dynamic_param(esRescaleV, EvshpId_tau_zE) = half
    dynamic_param(esRescaleV, EvshpId_brd_zE) = sqrt(half)

    !--- PROPS -------------------
    dynamic_prop(esSumType, :) = ESSumType_standard
    dynamic_prop(esSumType, EvshpId_brd_zE) = ESSumType_WithLbar_m

    dynamic_prop(esC1MuF, :) = esC1MuF_Q
  end subroutine dynamic_initiate
  
  !-------------------------------------------------------------------------
  ! Update a dynamic parameter (parameters are real, properties are integer?
  subroutine es_UpdateDynamicParam(handle, evshp_id, val)
    integer,  intent(in) :: handle, evshp_id
    real(dp), intent(in) :: val
    !--------------------------
    character(len=20) :: line
    call dynamic_initiate

    if (handle < lbound(dynamic_param,dim=1) .or. &
         & handle > ubound(dynamic_param,dim=1)) then
       write(line,*) handle
       call wae_Error('es_UpdateDynamicParam',&
            &'Unknown parameter handle '//trim(line))
    end if
    if (evshp_id < 1 .or. evshp_id > n_esid) then
       call wae_error('es_UpdateDynamicParam: illegal evshp_id')
    end if
    dynamic_param(handle, evshp_id) = val
  end subroutine es_UpdateDynamicParam

  !-------------------------------------------------------------------------
  ! Retrieve a dynamic parameter (parameters are real, properties are integer?
  subroutine es_GetDynamicParam(handle, evshp_id, val)
    integer,  intent(in)  :: handle, evshp_id
    real(dp), intent(out) :: val
    !--------------------------
    character(len=20) :: line
    call dynamic_initiate

    if (handle < lbound(dynamic_param,dim=1) .or. &
         & handle > ubound(dynamic_param,dim=1)) then
       write(line,*) handle
       call wae_Error('es_GetDynamicParam',&
            &'Unknown parameter handle '//trim(line))
    end if
    if (evshp_id < 1 .or. evshp_id > n_esid) then
       call wae_error('es_GetDynamicParam: illegal evshp_id')
    end if
    val = dynamic_param(handle, evshp_id) 
  end subroutine es_GetDynamicParam

  !-------------------------------------------------------------------------
  ! Update a dynamic property (parameters are real, properties are integer?
  subroutine es_UpdateDynamicProp(handle, evshp_id, val)
    integer,  intent(in) :: handle, evshp_id
    integer,  intent(in) :: val
    !--------------------------
    character(len=20) :: line
    call dynamic_initiate

    if (handle < lbound(dynamic_prop,dim=1) .or. &
         & handle > ubound(dynamic_prop,dim=1)) then
       write(line,*) handle
       call wae_Error('es_UpdateDynamicProp',&
            &'Unknown property handle '//trim(line))
    end if
    if (evshp_id < 1 .or. evshp_id > n_esid) then
       call wae_error('es_UpdateDynamicProp: illegal evshp_id')
    end if
    dynamic_prop(handle, evshp_id) = val
  end subroutine es_UpdateDynamicProp
  

  !-------------------------------------------------------------------------
  ! Retrieve a dynamic property (parameters are real, properties are integer?
  subroutine es_GetDynamicProp(handle, evshp_id, val)
    integer,  intent(in)  :: handle, evshp_id
    integer,  intent(out) :: val
    !--------------------------
    character(len=20) :: line
    call dynamic_initiate

    if (handle < lbound(dynamic_prop,dim=1) .or. &
         & handle > ubound(dynamic_prop,dim=1)) then
       write(line,*) handle
       call wae_Error('es_GetDynamicProp',&
            &'Unknown property handle '//trim(line))
    end if
    if (evshp_id < 1 .or. evshp_id > n_esid) then
       call wae_error('es_GetDynamicProp: illegal evshp_id')
    end if
    val = dynamic_prop(handle, evshp_id)
  end subroutine es_GetDynamicProp








  subroutine es_UpdateDynamicESMax(evshp_id, esmax_val)
    use warnings_and_errors
    integer,  intent(in)  :: evshp_id
    real(dp), intent(in)  :: esmax_val
    call es_UpdateDynamic(esMax, evshp_id, esmax_val)
    !call dynamic_initiate
    !if (evshp_id < 1 .or. evshp_id > size(dynamic_esmax)) then
    !   call wae_error('es_UpdateDynamicESMax: illegal evshp_id')
    !end if
    !dynamic_esmax(evshp_id) = esmax
  end subroutine es_UpdateDynamicESMax
 
  subroutine es_GetDynamicESMax(evshp_id, esmax_val)
    use warnings_and_errors
    integer,  intent(in)  :: evshp_id
    real(dp), intent(out) :: esmax_val
    call es_GetDynamic(esMax, evshp_id, esmax_val)
    !call dynamic_initiate
    !if (evshp_id < 1 .or. evshp_id > size(dynamic_esmax)) then
    !   call wae_error('es_GetDynamicESMax: illegal evshp_id')
    !end if
    !esmax = dynamic_esmax(evshp_id)
  end subroutine es_GetDynamicESMax
 
 
  subroutine es_UpdateDynamicESSumType(evshp_id, esSumType_val)
    use warnings_and_errors
    integer,  intent(in)  :: evshp_id
    integer,  intent(in)  :: esSumType_val
    call es_UpdateDynamic(esSumType, evshp_id, esSumType_val)
    !call dynamic_initiate
    !if (evshp_id < 1 .or. evshp_id > size(dynamic_esSumType)) then
    !   call wae_error('es_UpdateDynamicESSumType: illegal evshp_id')
    !end if
    !dynamic_esSumType(evshp_id) = esSumType
  end subroutine es_UpdateDynamicESSumType
 
  subroutine es_GetDynamicESSumType(evshp_id, esSumType_val)
    use warnings_and_errors
    integer,  intent(in)  :: evshp_id
    integer,  intent(out) :: esSumType_val
    call es_GetDynamic(esSumType, evshp_id, esSumType_val)
    !call dynamic_initiate
    !if (evshp_id < 1 .or. evshp_id > size(dynamic_esSumType)) then
    !   call wae_error('es_GetDynamicESSumType: illegal evshp_id')
    !end if
    !esSumType = dynamic_esSumType(evshp_id)
  end subroutine es_GetDynamicESSumType

  !======================================================================
  ! FROM HERE ONWARDS, ONLY PRIVATE ROUTINES
  !----------------------------------------------------------------------
  ! routine to decide which coefficient function gets called
  function es_Cgeneric(y, es_comp) result(res)
    use convolution_communicator
    real(dp), intent(in) :: y
    integer,  intent(in) :: es_comp
    real(dp)             :: res

    select case (es_id)
    case (EvshpId_tau_zQ)
       res = es_Cgeneric_tau_zQ(y, es_comp)
    case (EvshpId_tau_zE)
       res = es_Cgeneric_tau_zE(y, es_comp)
    case (EvshpId_rho_E, EvshpId_rho23E)
       res = es_Cgeneric_rho_E(y, es_comp)
    case (EvshpId_tau_tE)
       !-- constant pieces are identical for rho_E and tau_tE 
       !   (differences arise in rescale factor)
       res = es_Cgeneric_rho_E(y, es_comp)
    case (EvshpId_Cpr_E)
       !-- almost identical to rho_E (except for rescale factor)
       res = es_Cgeneric_rho_E(y, es_comp)
       !   but there is an additional delta-function piece
       !   (Mrinals mail of 6 Aug 2000)
       if (es_comp == es_comp_2q .and. cc_piece == cc_DELTA) then
          res = res + cf * pisq / 6.0_dp
          !res = res + cf * pisq / 8.0_dp
       end if
    case (EvshpId_brd_zE)
       res = es_Cgeneric_brd_zE(y, es_comp)
    case (EvshpId_Cempty)
       res = es_Cgeneric_Cempty(y, es_comp)
    case default
       write(0,*) 'ERROR. es_Cgeneric: unsupported event-shape&
            & variable: ', es_id
       stop
    end select
  end function es_Cgeneric
  

  !======================================================================
  ! eqns A.1 & A.2 of ADS
  function es_Cgeneric_tau_zQ(y, es_comp) result(res)
    use convolution_communicator
    use coefficient_functions
    real(dp), intent(in) :: y
    integer,  intent(in) :: es_comp
    real(dp)             :: res
    real(dp)             :: z
    real(dp) :: lnz

    z = exp(-y)
    res = zero

    select case(es_comp)
    case(es_comp_2q) !------------------------------------------------

       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          lnz = log(z)
          res = -CF*( -(1+z**2)/(1-z) * lnz + (1+2*z-6*z**2)/(two*(1-z)))
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
          res = res + CF*(-3/(two*(1-z)))
       case(cc_DELTA)
          res = zero
       end select
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_2g) !------------------------------------------------
       lnz = log(z)

       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          lnz = log(z)
          res = -TR*(-((1-z)**2+z**2)*lnz - (1-6*z*(1-z)))
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_Lq) !------------------------------------------------
       res = -cf_CqFL(y)
    case(es_comp_Lg) !------------------------------------------------
       res = -cf_CgFL(y)
    case default
       write(0,*) 'ERROR. es_Cgeneric_tau_zQ: unsupported es_comp =',es_comp
       stop
    end select
  end function es_Cgeneric_tau_zQ


  !======================================================================
  ! eqns A.5 & A.6 of ADS
  function es_Cgeneric_tau_zE(y, es_comp) result(res)
    use convolution_communicator
    use coefficient_functions
    real(dp), intent(in) :: y
    integer,  intent(in) :: es_comp
    real(dp)             :: res
    real(dp)             :: z
    real(dp) :: lnz, ln1mz
    !real(dp) :: fiddle

    z = exp(-y)
    res = zero

    select case(es_comp)
    case(es_comp_2q) !------------------------------------------------

       !fiddle = -two*ln2
       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          lnz = log(z); ln1mz = log(one - z)
          !-- NB: using new version of A.5, with ln(2(1-z)) -> ln(1-z)
          res = -CF*( -(1+z**2)/(1-z) * lnz &
               &      +(1+z**2)*(ln1mz)/(1-z) &
               &      +(1 + four*ln2 + 2*z - 6*z**2)/(two*(1-z)) &
               &      -(1+z)*ln2  )! + fiddle*(-(one+z)*half+one/(1-z)))
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
          lnz = log(z); ln1mz = log(one - z)
          !-- NB: using new version of A.5, with ln(2(1-z)) -> ln(1-z)
          res = res + CF*( +(2)*(ln1mz)/(1-z)&
               &      +(four*ln2 - 3)/(two*(1-z)))! +fiddle*(+one/(1-z)))
       case(cc_DELTA)
          res = -CF*( 1.5_dp*(ln2**2-ln2) )
       end select
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_2g) !------------------------------------------------
       lnz = log(z)

       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          lnz = log(z); ln1mz = log(one - z)
          res = -TR*( (z**2+ (1-z)**2)*(ln2 + ln1mz - lnz) &
               &- (1 - 6*z*(1-z)) )
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_Lq) !------------------------------------------------
       res = -cf_CqFL(y)
    case(es_comp_Lg) !------------------------------------------------
       res = -cf_CgFL(y)
    case default
       write(0,*) 'ERROR. es_Cgeneric_tau_zE: unsupported es_comp =',es_comp
       stop
    end select
  end function es_Cgeneric_tau_zE



  !======================================================================
  ! eqns A.5 & A.6 of ADS
  function es_Cgeneric_rho_E(y, es_comp) result(res)
    use convolution_communicator
    use coefficient_functions
    real(dp), intent(in) :: y
    integer,  intent(in) :: es_comp
    real(dp)             :: res
    real(dp)             :: z
    real(dp) :: lnz, ln1mz

    z = exp(-y)
    res = zero

    ! Using Mrinals equations of June 12th ===========================
    select case(es_comp)
    case(es_comp_2q) !------------------------------------------------
       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          if (z > half) then
             lnz = log(z); ln1mz = log(one - z)
             res = CF*( (half+6*z**3-5*z**2)/(1-z)&
                  &     + (1+z**2)*(ln1mz-lnz)/(1-z))
          end if
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
          ln1mz = log(one - z)
          res = res - CF*(1.5_dp/(1-z) + 2*ln1mz/(1-z))
       case(cc_DELTA)
          res = zero
       end select
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_2g) !------------------------------------------------

       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          if (z > half) then
             lnz = log(z); ln1mz = log(one - z)
             res = half*TR*((z**2+(1-z)**2)*(4*z-2+2*ln1mz-2*lnz)&
                  & +8*z*(1-z)*(1-2*z))
          end if
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_Lq) !------------------------------------------------
       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          if (z > half) then
             res = CF*(2*z-4*z**2)
          end if
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_Lg) !------------------------------------------------
       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          if (z > half) then
             res = half*TR*( 8*z*(1-z)*(1-2*z) )
          end if
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_CA2q) !------------------------------------------------
       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          if (z > half) then
             lnz = log(z)
             res = CF*(3*z**3 + 2 + z - 6*z**2 + (4*z-3-z**2)/(2*(1-z))&
                  &    - (1+z**2)/(1-z)*lnz )
          else
             ln1mz = log(one - z)
             res = CF*(3*z**3 +(2*z - 5*z**2)/(2*(1-z))&
                  &    - (1+z**2)/(1-z)*ln1mz )
          end if
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_CALq) !------------------------------------------------
       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          if (z > half) then
             res = CF*( 2*z*(1-z)**2 )
          else
             res = CF*( 2*z**3 )
          end if
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       if (cc_piece /= cc_DELTA) res = res * z
    case default
       write(0,*) 'ERROR. es_Cgeneric_rho_E: unsupported es_comp =',es_comp
       stop
    end select

    !-- need to subtract the likelyhood of there being nothing in the
    !   current hemisphere
    !   (not the quickest way of doing things?)
    res = res - es_Cgeneric_Cempty(y, es_comp)
  end function es_Cgeneric_rho_E
  

  !======================================================================
  ! As of 4/07/2001 these coefficient functions are still to be
  ! determined.
  function es_Cgeneric_brd_zE(y, es_comp) result(res)
    use convolution_communicator
    use coefficient_functions
    real(dp), intent(in) :: y
    integer,  intent(in) :: es_comp
    real(dp)             :: res
    real(dp)             :: z
    real(dp) :: lnz, ln1mz
    !real(dp) :: fiddle

    z = exp(-y)
    res = zero

    !--- THESE HAVE NOT YET BEEN WRITTEN 
    !    For the F2 piece they will be set to zero...
    !    REMEMBER TO RESET THEM...
    !    For the FL piece we do as for other vars

    select case(es_comp)
    case(es_comp_2q) !------------------------------------------------
       ! quark F2 piece

       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          lnz = log(z); ln1mz = log(one - z)
          res = CF*( (1+z**2)/(1-z)*lnz - (1+z**2)*ln1mz/(1-z) + &
               &     (6*z**2- 2*z - 1)/(2*(1-z)) )
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
          lnz = log(z); ln1mz = log(one - z)
          res = res + CF*(2*ln1mz/(1-z) - 1.5_dp/(1-z))
       case(cc_DELTA)
          res = -CF*(-3*ln2 + 2*ln2**2 + 3 - pisq/three)
       end select
       if (cc_piece /= cc_DELTA) res = res * z

    case(es_comp_2g) !------------------------------------------------
       ! gluon F2 piece

       lnz = log(z)

       select case(cc_piece)
       case(cc_REAL,cc_REALVIRT)
          lnz = log(z); ln1mz = log(one - z)
          res = -half*TR*(z**2+(1-z)**2)*(2*(ln1mz-lnz)-2)
       end select
       select case(cc_piece)
       case(cc_VIRT,cc_REALVIRT)
       case(cc_DELTA)
          res = zero
       end select
       
       if (cc_piece /= cc_DELTA) res = res * z
       !-- because what is given above is only the transverse piece
       res = res - cf_CgFL(y)

    case(es_comp_Lq) !------------------------------------------------
       res = -cf_CqFL(y)
    case(es_comp_Lg) !------------------------------------------------
       res = -cf_CgFL(y)
    case default
       write(0,*) 'ERROR. es_Cgeneric_brd_zE: unsupported es_comp =',es_comp
       stop
    end select
  end function es_Cgeneric_brd_zE

  !======================================================================
  ! for the time being 
  function es_Cgeneric_Cempty(y, es_comp) result(res)
    use convolution_communicator
    use coefficient_functions
    real(dp), intent(in) :: y
    integer,  intent(in) :: es_comp
    real(dp)             :: res
    real(dp)             :: z
    real(dp) :: lnz, ln1mz

    res = zero
    !-- unusual way of doing things: just real part
    if (cc_piece /= cc_REAL .and. cc_piece /= cc_REALVIRT) return
    

    z = exp(-y)
    if (z > half) return

    select case(es_comp)
    case(es_comp_2q) !------------------------------------------------
       ln1mz = log(one-z); lnz = log(z)
       res = CF*((6*z**3 + half - 5*z**2) + (1+z**2)*(ln1mz - lnz))/(1-z)
    case(es_comp_2g) !------------------------------------------------
       ln1mz = log(one-z); lnz = log(z)
       res = half*TR * ((z**2 + (1-z)**2)*( 4*z-2 - 2*lnz + 2*ln1mz)&
            & +8*z*(1-z)*(1-2*z))
    case(es_comp_Lq) !------------------------------------------------
       res = CF*2*z*(1-2*z)
    case(es_comp_Lg) !------------------------------------------------
       res = half*TR * (8*z*(1-z)*(1-2*z))
    case(es_comp_CA2q,es_comp_CALq)
       res = zero
    case default
       write(0,*) 'ERROR. &
            &es_Cgeneric_Cempty: unsupported es_comp =',es_comp
       stop
    end select
   
    if (cc_piece /= cc_DELTA) res = res * z

  end function es_Cgeneric_Cempty
  
    
end module event_shapes
!((x**2+(1-x)**2)*(4*x-2+2*log(1/x-1)) + 8*x*(1-x)*(1-2*x))



!======================================================================
! This is useful if one ONLY needs access to the event-shape id's
! Remember to update it when we add something to the list
!======================================================================
module event_shape_ids
  use event_shapes
  implicit none
  private
  
  public :: EvshpId_illegal
  public :: EvshpId_tau_zQ 
  public :: EvshpId_tau_zE 
  public :: EvshpId_tau_tQ 
  public :: EvshpId_tau_tE 
  public :: EvshpId_brd_zE 
  public :: EvshpId_brd_tE 
  public :: EvshpId_rho_E  
  public :: EvshpId_Cpr_E  
  public :: EvshpId_CrntEQ 
  public :: EvshpId_Cempty 
  public :: EvshpId_rho23E 

end module event_shape_ids
