! $Id: evolution.f90,v 1.5 2001/06/29 09:43:53 salam Exp $

module evolution
  use types; use consts_dp
  use conv_objects; use alfas_server
  use assertions; use qcd
  use warnings_and_errors
  implicit none
  private

  public :: ev_evolveLO, ev_evolveNLO, ev_evolve
  public :: ev_Setdt
  
  !-- fact that alpha_s/2pi is small means that we can quite comfortably
  !   take large steps
  !   have tested with 0.4 down to 1.5 GeV
  real(dp) :: dt_ballpark = 0.4
  !real(dp), parameter :: dt_ballpark = 0.2

  type(PMat),      pointer :: ev_PLO, ev_PNLO
  type(as_handle), pointer :: ev_ash
  integer                  :: ev_nloop
  real(dp) :: ev_muR_Q, fourpibeta0_lnmuR_Q



contains

  !-----------------------------------------------------
  ! At least give the user the option of setting this...
  subroutine ev_Setdt(dt)
    real(dp), intent(in) :: dt
    dt_ballpark = dt
  end subroutine ev_Setdt
  

  !======================================================================
  ! General entry routine for evolution. Takes a human distribution
  ! and deals with the necessary conversion to and from "evolution" format. 
  ! It works for LO and NLO evolution: the order being set by "nloop"
  subroutine ev_evolve(sh,pdfset, nloop, ash, Q_init, Q_end, muR_Q)
    use holders; use pdf_representation
    type(sigma_holder), intent(in)         :: sh
    real(dp),        intent(inout)         :: pdfset(:,:)
    integer,         intent(in)            :: nloop
    type(as_handle), intent(in), target    :: ash
    real(dp),        intent(in)            :: Q_init, Q_end
    real(dp),        intent(in), optional  :: muR_Q
    !-----------------------------------------------------
    real(dp) :: pdf_ev(size(pdfset,dim=1), size(pdfset,dim=2))
    
    !-- put things into the right format. NB NOT RELEVANT for disresum
    !call pdfr_HumanToEvln(sh%prep, pdfset, pdf_ev)

    select case(nloop)
    case(1)
       call ev_evolveLO(sh%P, pdf_ev, ash, Q_init, Q_end, muR_Q)
    case(2)
       call ev_evolveNLO(sh%P, sh%P1, pdf_ev, ash, Q_init, Q_end, muR_Q)
    case default
       call wae_error('ev_evolve: nloops other than 1 and 2 are not supported')
    end select

    !-- put things back into a "human" format. NB NOT RELEVANT for disresum
    !call pdfr_EvlnToHuman(sh%prep, pdf_ev, pdfset)
  end subroutine ev_evolve
  
  !======================================================================
  ! evolve the distribution pdfset with LO evolution from Q2_init to
  ! Q2_end, using the alpha_s defined by ash
  subroutine ev_evolveLO(PLO, pdfset, ash, Q_init, Q_end, muR_Q)
    use runge_kutta
    type(PMat),      intent(in), target    :: PLO
    real(dp),        intent(inout)         :: pdfset(:,:)
    type(as_handle), intent(in), target    :: ash
    real(dp),        intent(in)            :: Q_init, Q_end
    real(dp),        intent(in), optional  :: muR_Q
    !------------------------------------------------------------
    real(dp) :: t1, t2, dt, t
    integer  :: n, i

    ev_PLO => PLO
    ev_ash => ash
    ev_nloop = 1
    
    ev_muR_Q = default_or_opt(one,muR_Q)

    call ev_evolveLocal(pdfset, Q_init, Q_end)
  end subroutine ev_evolveLO



  !======================================================================
  ! evolve the distribution pdfset with LO evolution from Q2_init to
  ! Q2_end, using the alpha_s defined by ash
  subroutine ev_evolveNLO(PLO, PNLO, pdfset, ash, Q_init, Q_end, muR_Q)
    use runge_kutta
    type(PMat),      intent(in), target    :: PLO, PNLO
    real(dp),        intent(inout)         :: pdfset(:,:)
    type(as_handle), intent(in), target    :: ash
    real(dp),        intent(in)            :: Q_init, Q_end
    real(dp),        intent(in), optional  :: muR_Q
    !------------------------------------------------------------
    real(dp) :: t1, t2, dt, t
    integer  :: n, i

    ev_PLO  => PLO
    ev_PNLO => PNLO
    ev_ash => ash
    ev_nloop = 2

    ev_muR_Q = default_or_opt(one,muR_Q)

    call ev_evolveLocal(pdfset, Q_init, Q_end)
  end subroutine ev_evolveNLO


  !----------------------------------------------------------------------
  ! this bit here does the actual evolution. Could make it more
  ! sophisticated later, if necessary (e.g. variable step size).
  subroutine ev_evolveLocal(pdfset, Q_init, Q_end)
    use runge_kutta
    real(dp),        intent(inout)         :: pdfset(:,:)
    real(dp),        intent(in)            :: Q_init, Q_end
    !------------------------------------------------------------
    real(dp) :: t1, t2, dt, t
    integer  :: n, i

    fourpibeta0_lnmuR_Q = four*pi*beta0*log(ev_muR_Q)
    !write(0,*) fourpibeta0_lnmuR_Q
    t1 = two*log(Q_init)
    t2 = two*log(Q_end)

    if (t2 == t1) return

    n  = ceiling(abs(t2-t1)/dt_ballpark)
    dt = (t2 - t1)/n

    t = t1
    do i = 1, n
       call rkstp(dt, t, pdfset, ev_conv)
    end do
  end subroutine ev_evolveLocal


  !----------------------------------------------------------------------
  subroutine ev_conv(t, pdfset, dpdfset)
    real(dp), intent(in)  :: t, pdfset(:,:)
    real(dp), intent(out) :: dpdfset(:,:)
    !--------------------------------------
    real(dp) :: as2pi, Q
    type(PMat) :: Pfull

    Q       = exp(half*t)
    as2pi   = as_value(ev_ash, Q*ev_muR_Q)/twopi
    select case (ev_nloop)
    case(1)
       dpdfset = as2pi * (ev_PLO .conv. pdfset)
    case(2)
       if (fourpibeta0_lnmuR_Q /= zero) then
          call cobj_InitSplit(Pfull, ev_PLO, one + as2pi*fourpibeta0_lnmuR_Q)
       else
          call cobj_InitSplit(Pfull, ev_PLO)
       end if
       call cobj_AddSplit(Pfull, ev_PNLO, as2pi)
       dpdfset = as2pi * (Pfull .conv. pdfset)
       call cobj_DelSplit(Pfull)
    end select
  end subroutine ev_conv
  
  

end module evolution
