module mfm_helper
  use types; use consts_dp
  use driver_withdsm_hidden
  use resummation_definitions ! for powertype
  use event_shapes
  use alfas_server
  use sub_defs_io
  use warnings_and_errors
  implicit none
  private

  integer :: FOE_handle_lo, FOE_handle_hi, evshp_id
  
  !-- keep the results here?
  !   (Qlo,Qhi,Qmn,res,errst,errsp,errsp,errtot,thr) * 1:nbin * 0:nQ
  real(dp), allocatable :: exp_res(:,:), xQvals(:,:), Qvals(:)
  logical,  allocatable :: mask(:)
  integer :: nbin, nQ
  !-- these are the resummation properties
  integer :: PowerType
  real(dp) :: muR_Q, muI
  real(dp) :: Qmin, Qmax
  real(dp) :: LastAlfas, LastAlfa0, LastChi2=0
  real(dp) :: LastAsErr = zero, LastA0Err = zero
  integer  :: LastNpnt=0
  logical  :: include_systematics = .true.

  character(len=40) :: FOFile
  character(len=6) :: evshp_name

  public :: mfmh_Start, mfmh_GetThr, mfmh_Write, mfmh_WriteProp, mfmh_DoFit
  public :: mfmh_WriteCont

  integer, parameter :: ix_Q   = 3
  integer, parameter :: ix_res = 4
  integer, parameter :: ix_stt = 5
  integer, parameter :: ix_err = 8
  integer, parameter :: ix_thr = 9

contains

  !-- takes care of initialisation ---------------
  subroutine mfmh_Start
    integer  :: FOE_handle, iQ, idev, iostat, lcl_evshp_id
    real(dp) :: xbj, ybj,Q
    character(len=200) :: line
    evshp_name = string_val_opt('-evshp','tau_tE')
    write(0,*) 'Will attempt to do a fit for ',trim(evshp_name),' mean values'
    evshp_id = es_idOfName(trim(evshp_name))
    
    !-- use interface for the lazy ---------
    FOFile = string_val_opt('-FOFile', 'h1-fixeds.top')
    call dwdh_start(FOFile, evshp_id, FOE_handle_lo, FOE_handle_hi)

    !-- get the data ------------------------------
    idev = idev_open_opt('-datfile','H1-means.dat')
    nQ   = FOE_handle_hi - FOE_handle_lo
    allocate(exp_res(9,0:nQ))
    do
       read(idev,'(a)',iostat=iostat) line
       if (iostat /= 0) call wae_error('mfmh_Start',&
            &'Could not find requested event shape '//&
            &trim(evshp_name)//' in data file')
       if (index(line,trim(evshp_name)) > 0) then
          read(idev,'(a)') line    ! dummy line
          do iQ = 0, nQ
             read(idev,*,iostat=iostat) exp_res(1:7,iQ)
             if (iostat /= 0) call wae_error('mfmh_Start',&
                  &'Error on reading data')
          end do
          exit
       end if
    end do
    !-- get a reasonable approximation for the errors
    exp_res(8,:) = sqrt(exp_res(5,:)**2 &
         &              + (half*(exp_res(6,:)+exp_res(7,:)))**2 )
             

    allocate (xQvals(2, 0:nQ))
    do FOE_handle = FOE_handle_lo, FOE_handle_hi
       iQ = FOE_handle - FOE_handle_lo
       call dwdh_info(FOE_handle, xbj, ybj, Q, lcl_evshp_id)
       xQvals(:, iQ) = (/ xbj, Q /)
       if (log_val_opt('-printQvals')) &
            &write(0,*) 'bin ', FOE_handle - FOE_handle_lo, ': Q = ',Q 
       if (abs(xQvals(2, iQ) - exp_res(3,iQ)) > 0.01_dp) then 
          write(0,*) 'Warning, inconsistent Q values:', exp_res(3,iQ),&
               & xQvals(2, FOE_handle - FOE_handle_lo)
       end if
    end do

    !-- set up properties of resummation ------------------------------
    PowerType = PowerType_modshift
    muR_Q     = dble_val_opt('-muR',one)
    muI       = 2.0_dp

    !-- set up the mask ----
    allocate(mask(0:nQ),Qvals(0:nQ))
    Qvals = xQvals(2,:)
    Qmax = 100.0_dp
    Qmin =   5.0_dp

    !-- allow user reset
    Qmin = dble_val_opt('-Qmin',Qmin)
    Qmax = dble_val_opt('-Qmax',Qmax)

    include_systematics = log_val_opt('-syst', include_systematics)

    mask = (Qvals > Qmin) .and. (Qvals < Qmax)

    deallocate(Qvals)
  end subroutine mfmh_Start


   !----------------------------------------------------------
   ! The routine needed by Minuit; fval must contain chi^2
   subroutine mfmh_MinFunc(npar,grad,fval,xval,iflag,futil)
     integer,  intent(in)    :: npar, iflag
     real(dp), intent(inout) :: grad(npar), fval
     real(dp), intent(in)    :: xval(npar)
     real(dp), external      :: futil ! will not be used
     !---------------------------------------------------
     real(dp) :: chi2(size(mask,dim=1))
     integer  :: ix_lcl_err
     
     if (include_systematics) then
        ix_lcl_err = ix_err
     else
        ix_lcl_err = ix_stt
     end if

     ! generate the distribution
     call mfmh_GetThr(xval(1),xval(2))
     ! work out the chi^2
     where (mask)
        chi2 = ((exp_res(ix_res,:) - &
             &      exp_res(ix_thr,:))/exp_res(ix_lcl_err,:))**2
     elsewhere
        chi2 = zero
     end where
     fval = sum(chi2)
     !write(0,*) xval(1:2), fval
   end subroutine mfmh_MinFunc
   
   !------------------------------------------------------
   ! Run the fitting software?
   subroutine mfmh_DoFit
     integer :: unit_scr = 10, ierflag
     integer :: i
     character(len=10) :: parname
     real(dp)          :: val(2), err(2), bnd1, bnd2
     real(dp)          :: chi2, chi2_dist, errdef
     integer           :: ni, nx, istat
     integer           :: count(size(mask,dim=1))
     call mninit(5,unit_scr,7)
     !call mnparm(1, "alfasMZ", 0.118_dp, 0.002_dp, zero, zero, ierflag)
     call mnparm(1, "alfasMZ", 0.118_dp, 0.002_dp, 0.05_dp, 0.15_dp, ierflag)
     call mnparm(2, "alfa0",   0.500_dp, 0.050_dp, zero, zero, ierflag)
 
     !-- fix alfas
     if (log_val_opt('-fixas')) then
        val(1) = dble_val_opt('-fixas')
        call mnparm(1, "alfasMZ", val(1), &
             &0.002_dp, zero, zero, ierflag) 
        call mncomd(mfmh_MinFunc,'fix 1',ierflag,0)
        write(0,*) 'Fixing alfasMZ = ', val(1)
     end if
     
     if (log_val_opt('-fixa0')) then
        val(2) = dble_val_opt('-fixa0')
        call mnparm(2, "alfa0", val(2), 0.050_dp, zero, zero, ierflag) 
        call mncomd(mfmh_MinFunc,'fix 2',ierflag,0)
        write(0,*) 'Fixing alfa0 = ', val(2)
     end if
     
     call mncomd(mfmh_MinFunc, "minimize", ierflag, 0)
     do i = 1, 2
        call mnpout(i, parname, val(i), err(i), bnd1, bnd2, ierflag)
        write(0, '(a,f7.4," +-",f7.4)') parname, val(i), err(i)
     end do
     call mnstat(chi2, chi2_dist,errdef,ni,nx,istat)
     where (mask)
        count = 1
     elsewhere
        count = 0
     end where
 
     LastAsErr = err(1)
     LastA0Err = err(2)
     LastChi2 = chi2; LastNpnt = sum(count)
     write(0,'("chi2 = ",f8.2," for",i4," points")') chi2, LastNpnt
     !-- make sure we have the right distribution
     call mfmh_GetThr(val(1),val(2))
   end subroutine mfmh_DoFit
   
   !----------------------------------------------------------
   ! Writes all the details of the fit, including a leading #
   ! so that it can be part of a gnuplot file.
   subroutine mfmh_WriteProp(idev)
     integer, intent(in) :: idev
     integer  :: ival
     real(dp) :: dval
 
     write(idev,'("# alfas(MZ): ",f8.6," +- ",f8.6)') LastAlfas, LastAsErr
     write(idev,'("# alfa0:     ",f8.6," +- ",f8.6)') LastAlfa0, LastA0Err
     write(idev,'("# chi2 = ",f8.2," for",i4," points")') LastChi2, LastNpnt
 
     write(idev,'("#")')
     write(idev,'("# muR_Q:     ",f8.6)') muR_Q
     write(idev,'("# muI:       ",f8.6)') muI
 
     write(idev,'("#")')
     write(idev,'("# FOFile:    ",a)') FOFile
     write(idev,'("# Evshp:     ",a)') evshp_name
     
     write(idev,'("#")')
     write(idev,'("# PowerType:     ",i2)') PowerType
     write(idev,'("# Include systematics: ",l2)') include_systematics
 
     write(idev,'("#")')
     write(idev,'("# Qmin:      ",f10.6)') Qmin
     write(idev,'("# Qmax:      ",f10.6)') Qmax
     
   end subroutine mfmh_WriteProp
   
   !-------------------------------------------------------------
   ! write out the results of the analysis
   subroutine mfmh_Write(idev)
     integer, intent(in) :: idev
     integer :: iQ, i
     real(dp) :: Qfact = 10.0_dp
     do iQ = 0, nQ
        !useit = .not.(Qval < 10.0_dp .or. exp_res(2,i,iQ) < 2.0_dp/Qval &
        !     & .or. exp_res(2,i,iQ) > 0.31_dp)
        !if (.not. useit) cycle
        write(idev,'(9es14.5,l)') exp_res(:,iQ)
        !write(idev,'(5es14.5,l)') exp_res(1:2,i,iQ), &
        !     &exp_res(3:5,i,iQ)*Qfact**(nQ-iQ), mask(i,iQ)
     end do
   end subroutine mfmh_Write
   
   !-------------------------------------------------------------
   ! write out a 1-sigma error contour 
   subroutine mfmh_WriteCont(idev)
     integer, intent(in) :: idev
     !------------------------------
     integer, parameter :: npnt = 100
     integer :: i, imod, ierflag, nfound
     real(dp) :: chi2, asmz(npnt), a0c(npnt), nstd2(1)
     nstd2 = one
     call mnexcm(chi2,"set err",nstd2,1,ierflag,0)
     call mncont(mfmh_MinFunc, 1, 2, npnt, asmz, a0c, nfound, 0)
     !-- tricks
     do i = npnt/4, npnt+npnt/4
        imod = mod(i-1,npnt-1)+1
        write(idev,*) asmz(imod), a0c(imod)
     end do
     write(idev,*)
     write(idev,*)

     nstd2 = four
     call mnexcm(chi2,"set err",nstd2,1,ierflag,0)
     call mncont(mfmh_MinFunc, 1, 2, npnt, asmz, a0c, nfound, 0)
     !-- tricks
     do i = npnt/4, npnt+npnt/4
        imod = mod(i-1,npnt-1)+1
        write(idev,*) asmz(imod), a0c(imod)
     end do
 
     !-- try again
     nstd2 = one
     call mnexcm(chi2,"set err",nstd2,1,ierflag,0)
   end subroutine mfmh_WriteCont
   
 
   !-------------------------------------------------------------
   ! set up the theory prediction in entry 5 of the exp_res array
   subroutine mfmh_GetThr(alfas_MZ, alfa0)
     real(dp), intent(in) :: alfas_MZ, alfa0
     !------------------------------------
     type(as_handle)   :: ash
     real(dp) :: alfas
     integer  :: iQ, FOE_handle, i
     real(dp) :: Qval
 
     LastAlfas = alfas_MZ
     LastAlfa0 = alfa0
 
     call as_init(ash, alfas_MZ)
 
     do FOE_handle = FOE_handle_lo, FOE_handle_hi
        iQ = FOE_handle - FOE_handle_lo
        exp_res(ix_thr,iQ) = dwdh_mean_ash(FOE_handle, ash, muR_Q,&
                                               & alfa0, muI, PowerType)
     end do
     
   end subroutine mfmh_GetThr

end module mfm_helper


program main_fitdist
  use mfm_helper; use types; use sub_defs_io
  implicit none
  logical :: checkallargs, docontour 
  integer :: idev, idev_cont

  call mfmh_start

  if (log_val_opt('-outbase')) then
     idev = idev_open_opt('-outbase',ext='.res')
  else
     idev = 6
  end if

  docontour = log_val_opt('-cont')
  

  if (log_val_opt('-fit',.true.)) then
     call mfmh_DoFit
  else
     call mfmh_GetThr(0.118_dp, 0.5_dp)
     write(0,*) 'Not doing a fit'
  end if

  call mfmh_WriteProp(idev)
  call mfmh_Write(idev)

  if (docontour) then
     idev_cont = idev_open_opt('-outbase',ext='.cnt')
     call mfmh_WriteCont(idev_cont)
  end if

  checkallargs = CheckAllArgsUsed(0)
  
end program main_fitdist
