!======================================================================
! Module for tabulating pdf at a given x value for a range of Q values. 
! 
! For the time being support only the most primitve version, one that
! generates table directly from the pdf routine.
!
!
! PQTB == pdf table
!
! $Id: pdf_tabulate.f90,v 1.11 2001/08/20 18:24:02 salam Exp $
! GPS 20/03/2001 (CCN19-56)
module pdf_tabulate
  use types; use consts_dp
  use parton_distributions; use pdf_representation
  use interpolation; use holders; use evolution
  use alfas_server; use convolution; use conv_objects
  implicit none
  private

  !-- from PRM: directly from MRS, CTEQ etc...
  integer, parameter, public :: qtabid_illegal = -1
  integer, parameter, public :: qtabid_fromPRM = 1
  integer, parameter, public :: qtabid_evlnLO  = 2
  integer, parameter, public :: qtabid_evlnNLO = 3

  
  integer,  parameter :: undefined_intval = -1000000034
  real(dp), parameter :: undefined_dpval  = -1e90_dp

  type qtab_holder
     real(dp), pointer :: f2lo(:), gamma(:), lnQ(:)
     real(dp), pointer :: C2L(:), C2L_gamma(:)
     real(dp)          :: xval, yval, muF, dlnQ, alfas_muR, muR, muR_Q
     type(as_handle)   :: ash
     integer           :: qtabid, pdf_id, es_id, n
     !------ for use by cache --------------
     type(qtab_holder), pointer :: parent
     logical                    :: proxy
  end type qtab_holder

  !---- need this if doing fits for certain variables ----
  !     it should enable significant speeding up of procedure
  type qtab_cache
     type(qtab_holder), pointer :: qth(:)
     integer,           pointer :: access_order(:)
     real(dp)                   :: Qlo
     integer                    :: n
  end type qtab_cache
  

  interface qtab_InitQTAB
     module procedure qtab_InitQTAB_fromPRM, qtab_InitQTAB_evln
  end interface

  public :: qtab_holder, qtab_cache
  public :: qtab_InitQTAB, qtab_Val, qtab_Gamma, qtab_C2L, qtab_DelQTAB
  public :: qtab_InitCache, qtab_GetQTH
  
  real, parameter :: dlnQ_ballpark = 0.2_dp

contains

  

  !--------------------------------------------------------------------
  ! set up a table containin f2 leading order over the desired range of
  ! Q2 values.
  ! Do it based on a pre-existing grid-parton distribution at muF, which
  ! is evolved down according to some scheme specified by qtabid.
  !
  ! WARNING: alfas_muR must be evaluated at alfas(muR)
  subroutine qtab_InitQTAB_evln(qtabh, xval, Qlo, muF, &
       &qtabid, sh, qdist, alfas_muR, muR, muR_Q, esh, yval)
    use qcd; use assertions; use warnings_and_errors
    type(qtab_holder),  intent(out) :: qtabh
    real(dp),           intent(in)  :: xval, Qlo, muF
    integer,            intent(in)  :: qtabid
    type(sigma_holder), intent(in)  :: sh
    real(dp),           intent(in)  :: qdist(:,:), alfas_muR, muR
    real(dp), optional, intent(in)  :: muR_Q
    type(eventshape_holder), optional, intent(in) :: esh
    real(dp),                optional, intent(in) :: yval
    !----------------------------------------------------------
    real(dp), pointer :: qd(:,:), gamma(:,:)
    real(dp) :: lnmuF, lnQlo, lnQ, Qhere, Qlast, muR_Q_lcl
    real(dp) :: lnmuF_Q
    integer  :: i, n

    muR_Q_lcl = default_or_opt(one, muR_Q)

    !-- SHOULD WE WARN ABOUT non-standard cf? Affects beta1.
    if (ca /= three .or. tr /= half) call wae_error(&
         &'qtab_InitQTAB_evln:', &
         &'Non-standard cf,tr colour factors not supported')

    !write(0,*) 'qtab: ',muF, Qlo
    lnmuF = log(muF); lnQlo = log(Qlo)
    n = nint((lnmuF - lnQlo)/dlnQ_ballpark)
    qtabh%dlnQ      = (lnmuF - lnQlo) / n
    qtabh%n         = n
    qtabh%qtabid    = qtabid
    qtabh%alfas_muR = alfas_muR
    qtabh%muR_Q     = muR_Q_lcl
    qtabh%xval      = xval
    !-- this is implicitly contained in the array lnmuF, but it can 
    !   be useful to have it even if the array is not allocated
    qtabh%muF    = muF 
    qtabh%proxy  = .false.
    if (present(esh)) then
       qtabh%es_id = esh%id
       if (.not. present(yval)) call wae_error('qtab_InitQTAB_evln'&
            &'yval not supplied; but required when esh is supplied')
       qtabh%yval  = yval
    else
       qtabh%es_id = undefined_intval
       qtabh%yval  = undefined_dpval
    end if

    allocate(qtabh%f2lo(0:n), qtabh%gamma(0:n), qtabh%lnQ(0:n))
    call pdf_InitDist(sh%gd, qd, muF, pdf_ZERO)
    call pdf_InitDist(sh%gd, gamma, muF, pdf_ZERO)
    qd = qdist

    if (present(esh)) then
       allocate(qtabh%C2L(0:n), qtabh%C2L_gamma(0:n))
    else
       nullify(qtabh%C2L, qtabh%C2L_gamma)
    end if
    
    !-- actually need muF_Q to get right constant term
    lnmuF_Q = log(muF/muR * muR_Q_lcl)

    Qlast = muF
    select case(qtabid)
    case(qtabid_evlnLO)
       !call as_init(qtabh%ash, alfas_muR, muF*muR_Q_lcl,nloop=1,fixnf=nf_int)
       !-------------------------------------------------------------
       ! NB: CHANGE in logic after "v 1.5 2001/05/01 13:46:40"
       ! We assume muF = muF; since LO is intended to contain ONLY 
       ! double and single logs, the consistent approach is here to use
       ! alfas(mu) for derivative of structure function at scale muF.
       ! The following skullduggery should lead to that being satisfied
       call as_init(qtabh%ash, alfas_muR, muF*muR_Q_lcl, nloop=1, fixnf=nf_int)
       do i = n, 0,-1
          qtabh%lnQ(i)  = lnQlo + i * qtabh%dlnQ
          Qhere = exp(qtabh%lnQ(i))
          if (i /= n) call ev_evolveLO(sh%P, qd, qtabh%ash, &
               &Qlast, Qhere, muR_Q_lcl)
          qtabh%f2lo(i) = (sh%C2 .conv. qd) .atx. (xval .with. sh%gd)
          gamma = (sh%P .conv. qd)
          qtabh%gamma(i) = ((sh%C2 .conv. gamma) .atx. (xval .with. sh%gd)) &
               & / qtabh%f2lo(i)
          if (present(esh)) then
             !-- remember to include muF dependence, CCN20-94
             qtabh%C2L(i) = cobj_Eval2LConv(esh%C2_1, &
                  &esh%CL_1, qd, xval,yval) / qtabh%f2lo(i) &
                  &-two*lnmuF_Q*qtabh%gamma(i)
             qtabh%C2L_gamma(i) = cobj_Eval2LConv(esh%C2_1, &
                  &esh%CL_1, gamma, xval,yval) / qtabh%f2lo(i)&
                  &-two*lnmuF_Q*((sh%C2 .conv. (sh%P.conv.gamma)) &
                  &       .atx. (xval .with. sh%gd))/qtabh%f2lo(i)
          end if
          
          Qlast = Qhere
       end do
    case(qtabid_evlnNLO)
       call as_init(qtabh%ash, alfas_muR, muR, nloop=2, fixnf=nf_int)
       do i = n, 0,-1
          qtabh%lnQ(i)  = lnQlo + i * qtabh%dlnQ
          Qhere = exp(qtabh%lnQ(i))
          if (i /= n) call ev_evolveNLO(sh%P, sh%P1, qd, qtabh%ash, &
               &Qlast, Qhere, muR_Q_lcl)
          qtabh%f2lo(i) = (sh%C2 .conv. qd) .atx. (xval .with. sh%gd)
          gamma = (sh%P .conv. qd)
          qtabh%gamma(i) = &
               &((sh%C2 .conv. gamma) .atx. (xval .with. sh%gd))&
               & / qtabh%f2lo(i)
          if (present(esh)) then
             !-- remember to include muF dependence, CCN20-94
             qtabh%C2L(i) = cobj_Eval2LConv(esh%C2_1, &
                  &esh%CL_1, qd, xval,yval) / qtabh%f2lo(i) &
                  &-two*lnmuF_Q*qtabh%gamma(i)
             qtabh%C2L_gamma(i) = cobj_Eval2LConv(esh%C2_1, &
                  &esh%CL_1, gamma, xval,yval) / qtabh%f2lo(i)&
                  &-two*lnmuF_Q*((sh%C2 .conv. (sh%P.conv.gamma)) &
                  &       .atx. (xval .with. sh%gd))/qtabh%f2lo(i)
          end if
          Qlast = Qhere
       end do
    case default
       stop 'ERROR in qtab_InitQTAB_evln: unsupported qtabid'
    end select
    

    call conv_DelGridQuant(qd)
    call conv_DelGridQuant(gamma)
  end subroutine qtab_InitQTAB_evln
  

  !--------------------------------------------------------------------
  ! set up a table containin f2 leading order over the desired range of
  ! Q2 values 
  subroutine qtab_InitQTAB_fromPRM(qtabh, xval, Qlo, muF, iset)
    use qcd; use warnings_and_errors
    type(qtab_holder), intent(out) :: qtabh
    real(dp),         intent(in)   :: xval, Qlo, muF
    integer,          intent(in)   :: iset
    !-----------------------------------------------
    real(dp) :: lnmuF, lnQlo, lnQ, res(0:ncomponents)
    integer  :: i, n

    if (ca /= three .or. tr /= half) call wae_error(&
         &'qtab_InitQTAB_evln:', &
         &'Non-standard cf,tr colour factors not supported')

    lnmuF = log(muF); lnQlo = log(Qlo)
    n = nint((lnmuF - lnQlo)/dlnQ_ballpark)
    qtabh%dlnQ   = (lnmuF - lnQlo) / n
    qtabh%n      = n
    qtabh%qtabid = qtabid_fromPRM
    qtabh%pdf_id = iset
    qtabh%xval   = xval
    !-- this is implicitly contained in the array lnmuF, but it can 
    !   be useful to have it even if the array is not allocated
    qtabh%muF    = muF 
    qtabh%proxy  = .false.
    qtabh%es_id  = undefined_intval
    qtabh%yval   = undefined_dpval

    allocate(qtabh%f2lo(0:n), qtabh%lnQ(0:n))
    !-- NOT SUPPORTED IN THIS CASE, though they could be
    !   supported if it were needed!
    nullify(qtabh%gamma, qtabh%C2L, qtabh%C2L_gamma)

    do i = 0, n
       qtabh%lnQ(i)  = lnQlo + i * qtabh%dlnQ
       res           = pdf_Val(xval, exp(qtabh%lnQ(i)), iset)
       qtabh%f2lo(i) = (four * res(id_u) + res(id_d))/9.0_dp
    end do
    
  end subroutine qtab_InitQTAB_fromPRM
  

  !----------------------------------------------------------------------
  ! Given a table set up as before, return a value for f2 (leading order) 
  ! at the desired Q value. It will give a warning if one is asking for
  ! something outside the available range of Q
  function qtab_Val(qtabh, Qval) result(res)
    type(qtab_holder), intent(in), target  :: qtabh
    real(dp),          intent(in)  :: Qval
    real(dp)                       :: res
    !------------------------------------------------
    integer, parameter :: npnt = 4
    real(dp) :: dy, lnQ
    type(qtab_holder), pointer :: QTH_lcl, qtabh_p

    qtabh_p => qtabh
    call qtab_SelectQTH(qtabh_p, Qval, QTH_lcl, lnQ)

    call interpolate_uniform(QTH_lcl%lnQ, QTH_lcl%f2lo, lnQ, res, dy, npnt)
    !write(0,*) Qval, lnQ, res
  end function qtab_Val



  !----------------------------------------------------------------------
  ! Given a table set up as before, return a value for 
  ! (C2 .conv. PLO .conv. qd)/(C2 .conv. qd) at the desired Q value.
  ! As above, it will give a warning if one is asking for
  ! something outside the available range of Q
  function qtab_Gamma(qtabh, Qval) result(res)
    use warnings_and_errors
    type(qtab_holder), intent(in), target  :: qtabh
    real(dp),          intent(in)  :: Qval
    real(dp)                       :: res
    !------------------------------------------------
    integer, parameter :: npnt = 4
    real(dp) :: dy, lnQ
    type(qtab_holder), pointer :: QTH_lcl, qtabh_p

    qtabh_p => qtabh
    call qtab_SelectQTH(qtabh_p, Qval, QTH_lcl, lnQ)

    !-- if it is not associated then it probably means that we have
    !   qtabid_fromPRM, for which gamma is not currently supported
    !   (requires several program modifications)
    if (.not. associated(QTH_lcl%gamma)) call wae_error('qtab_Gamma',&
         &'Access to Gamma is supported only with evlnLO and &
         &evlnNLO, but not fromPRM')

    call interpolate_uniform(QTH_lcl%lnQ, QTH_lcl%gamma, lnQ, res, dy, npnt)
    !write(0,*) Qval, lnQ, res
  end function qtab_Gamma


  !----------------------------------------------------------------------
  ! Given a table set up as before, return values for 
  ! (C2L .conv. qd)/(C2 .conv. qd)
  ! (C2L .conv. PLO .conv. qd)/(C2 .conv. qd) at the desired Q value.
  ! As above, it will give a warning if one is asking for
  ! something outside the available range of Q
  subroutine qtab_C2L(qtabh, Qval, C2L, C2L_gamma)
    use warnings_and_errors
    type(qtab_holder), intent(in), target  :: qtabh
    real(dp),          intent(in)   :: Qval
    real(dp),          intent(out)  :: C2L, C2L_gamma
    !------------------------------------------------
    integer, parameter :: npnt = 4
    real(dp) :: dy, lnQ
    type(qtab_holder), pointer :: QTH_lcl, qtabh_p

    qtabh_p => qtabh
    call qtab_SelectQTH(qtabh_p, Qval, QTH_lcl, lnQ)

    !-- if it is not associated then it probably means that we have
    !   qtabid_fromPRM, for which gamma is not currently supported
    !   (requires several program modifications)
     if (.not. (associated(QTH_lcl%C2L) .and. &
         &  associated(QTH_lcl%C2L_gamma))) call wae_error('qtab_C2L',&
         &'Access to C2L is not supported with this qtabh.',&
         &'Perhaps because it uses qtabid_fromPRM')

    call interpolate_uniform(QTH_lcl%lnQ, QTH_lcl%C2L, lnQ, C2L, dy, npnt)
    call interpolate_uniform(QTH_lcl%lnQ, QTH_lcl%C2L_gamma, &
                                             & lnQ, C2L_gamma, dy, npnt)
  end subroutine qtab_C2L
  
  
  !------------------------------------------------------------------------
  ! select lnQ value and set up a pointer to the QTH that needs to be used. 
  ! (taking into account proxies and such things)
  subroutine qtab_SelectQTH(qtabh_p, Qval, QTH_lcl, lnQ)
    use warnings_and_errors
    type(qtab_holder), pointer     :: qtabh_p, QTH_lcl
    real(dp),          intent(in)  :: Qval
    real(dp),          intent(out) :: lnQ
    !--------------------------------------
    integer, parameter :: max_warn = 4
    !integer, parameter :: max_warn = 20
    integer, save      :: warn_id_lo = 0, warn_id_hi = 0
    character(len=14) :: qval_string

    if (qtabh_p%proxy) then
       !-- see CCN19-65 -----
       ! Just a single log effect, so purely a function of (as L)
       if (qtabh_p%qtabid /= qtabid_fromPRM) then
          lnQ = log(Qval/qtabh_p%muF) *&
               & qtabh_p%alfas_muR / qtabh_p%parent%alfas_muR &
               &+ log(qtabh_p%muF)
       else
          lnQ = log(Qval)
       end if
       QTH_lcl => qtabh_p%parent
       !-- MODIFY THINGS IF INTRODUCE MUR.
    else
       !-- use as was originally intended
       lnQ = log(Qval)
       QTH_lcl => qtabh_p
    end if
    
    if (lnQ < QTH_lcl%lnQ(0)) then
       write(qval_string,'(f14.5)') Qval
       call wae_warn(max_warn, warn_id_lo,&
         &'qtab_Val: requested Qval below lower limit:'//qval_string)
    end if
    if (lnQ > QTH_lcl%lnQ(QTH_lcl%n)) call wae_warn(max_warn, warn_id_lo,&
         &'qtab_Val: requested Qval above upper limit')
  end subroutine qtab_SelectQTH
  

  !-------------------------------------------------------------------
  ! We may want to clean up!
  ! But this had better not be used with something that has not been 
  ! initialised. 
  subroutine qtab_DelQTAB(qtabh)
    type(qtab_holder), intent(inout) :: qtabh
    if (.not. qtabh%proxy) then
       if (associated(qtabh%f2lo))  deallocate(qtabh%f2lo)
       if (associated(qtabh%gamma)) deallocate(qtabh%gamma)
       if (associated(qtabh%lnQ))   deallocate(qtabh%lnQ)
       if (associated(qtabh%C2L))   then
          deallocate(qtabh%C2L)
       end if
       
       if (associated(qtabh%C2L_gamma))  deallocate(qtabh%C2L_gamma)
    else
       nullify(qtabh%f2lo, qtabh%gamma, qtabh%lnQ, qtabh%parent,&
            &qtabh%C2L, qtabh%C2L_gamma)
    end if
  end subroutine qtab_DelQTAB
  






  !======================================================================
  !          FROM HERE ONE, STUFF RELATING DIRECTLY TO THE CACHE
  !======================================================================



  !----------------------------------------------------------------------
  ! Initialise a cache of the given size.
  !
  ! It is responsible for making sure that each element of the cache is
  ! "clean" when it starts.
  !
  ! Bear in mind that at start it will need to go through the whole
  ! of the cache in order to know if something is around, so one musn't 
  ! make the cache too large? Or just make it a little cleverer?
  subroutine qtab_InitCache(qtcache, Qlo, size)
    type(qtab_cache), intent(out) :: qtcache
    real(dp),         intent(in)  :: Qlo
    integer,          intent(in)  :: size
    !-------------------------------------
    integer :: i
    allocate(qtcache%qth(size), qtcache%access_order(size))
    qtcache%Qlo = Qlo
    qtcache%n = size
    do i = 1, qtcache%n
       qtcache%access_order(i) = i
       nullify(qtcache%qth(i)%parent, qtcache%qth(i)%f2lo, &
            &qtcache%qth(i)%gamma, qtcache%qth(i)%lnQ,&
            &qtcache%qth(i)%C2L, qtcache%qth(i)%C2L_gamma)
    end do
    !-- these are the things which may be checked...
    qtcache%qth%qtabid = qtabid_illegal
    qtcache%qth%pdf_id = pdf_ILLEGAL
    qtcache%qth%muF    = -one
    qtcache%qth%xval   = -one
  end subroutine qtab_InitCache

  !-----------------------------------------------------------------------
  ! Function to determine if two qtabhs are similar (i.e. whether we can 
  ! recycle one to give the other). 
  !
  ! Will need to add mu_R support later on
  function qtab_similar(qtabh1, qtabh2) result(res)
    type(qtab_holder),  intent(in) :: qtabh1, qtabh2
    logical                        :: res

    res = (qtabh1%muF    == qtabh2%muF) .and. &
         &(qtabh1%pdf_id == qtabh2%pdf_id) .and.&
         &(qtabh1%qtabid == qtabh2%qtabid) .and.&
         &(qtabh1%xval   == qtabh2%xval) .and.&
         &(qtabh1%es_id  == qtabh2%es_id) .and.&
         &(qtabh1%yval   == qtabh2%yval)

    !write(0,*) qtabh1%muF    == qtabh2%muF, qtabh1%pdf_id == qtabh2%pdf_id, &
    !     &qtabh1%qtabid == qtabh2%qtabid, qtabh1%xval   == qtabh2%xval
    
    ! strictly the above is slight overkill: if one is requesting
    ! something without any specific event shape one could still
    ! use a parameterisation with an event shape...
    
    if (.not. res) return
    select case(qtabh1%qtabid)
    case(qtabid_fromPRM, qtabid_evlnLO)
       res = .true.
    case(qtabid_evlnNLO)
       res = (qtabh1%alfas_muR == qtabh2%alfas_muR  .and. &
            & qtabh1%muR       == qtabh2%muR        .and. &
            & qtabh1%muR_Q     == qtabh2%muR_Q)
    case default
       res = .false.
    end select
  end function qtab_similar


  !------------------------------------------------------------------
  ! Looks for an appropriate entry in the cache. Generates it
  ! if necessary
  subroutine qtab_GetQTH(qtcache, qth, xval, muF, alfas_muR, muR, muR_Q,&
       & qtabid, pdf_id, sh, qdist, esh, yval)
    use warnings_and_errors
    type(qtab_cache),   intent(inout) :: qtcache
    type(qtab_holder),  intent(out)   :: qth
    real(dp),           intent(in)    :: xval, muF, alfas_muR, muR, muR_Q
    integer,            intent(in)    :: qtabid, pdf_id
    type(sigma_holder), intent(in)    :: sh
    real(dp),           intent(in)    :: qdist(:,:)
    type(eventshape_holder), optional, intent(in) :: esh
    real(dp),                optional, intent(in) :: yval
    !-----------------------------------------------------
    integer :: i, iqt
    logical :: similar

    !-- first "define" the qth
    qth%qtabid    = qtabid
    qth%xval      = xval
    qth%muF       = muF
    qth%alfas_muR = alfas_muR
    qth%muR       = muR
    qth%muR_Q     = muR_Q
    qth%pdf_id    = pdf_id
    qth%proxy     = .true.
    if (present(esh)) then
       qth%es_id = esh%id
       if (.not. present(yval)) call wae_error('qtab_GetQTH'&
            &'yval not supplied; but required when esh is supplied')
       qth%yval  = yval
    else
       qth%es_id = undefined_intval
       qth%yval  = undefined_dpval
    end if

    do i = 1, qtcache%n
       iqt = qtcache%access_order(i)
       similar = qtab_similar(qth, qtcache%qth(iqt))
       !write(0,*) i, iqt, similar, qth%muF, qtcache%qth(iqt)%muF
       if (qtab_similar(qth, qtcache%qth(iqt))) exit
    end do

    if (i <= qtcache%n) then
       qtcache%access_order(2:i) = qtcache%access_order(1:i-1)
       qtcache%access_order(1)   = iqt
    else
       !-- a right circular shift ---------------------------
       ! throws away everything least-used element
       qtcache%access_order = cshift(qtcache%access_order, -1)
       iqt = qtcache%access_order(1)
       !write(0,*) 'h',qtcache%access_order
       !-- now reinitialise a given element
       call qtab_DelQTAB(qtcache%qth(iqt))
       select case(qtabid)
       case(qtabid_fromPRM)
          call qtab_InitQTAB(qtcache%qth(iqt), xval, qtcache%Qlo, muF, pdf_id)
          ! alfas_muR not set, but not needed, so leave it alone
       case default
          call qtab_InitQTAB(qtcache%qth(iqt), xval, qtcache%Qlo, muF, &
               & qtabid, sh, qdist, alfas_muR, muR, muR_Q, esh, yval)
          !-- does not get set, so do it by hand (and it IS needed)
          qtcache%qth(iqt)%pdf_id = pdf_id
       end select
    end if
    
    qth%parent => qtcache%qth(iqt)
    
  end subroutine qtab_GetQTH











end module pdf_tabulate
